# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - EventAggregator library

    @copyright: 2008, 2009, 2010, 2011 by Paul Boddie <paul@boddie.org.uk>
    @copyright: 2000-2004 Juergen Hermann <jh@web.de>,
                2005-2008 MoinMoin:ThomasWaldmann.
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin.Page import Page
from MoinMoin.action import cache
from MoinMoin import caching
from MoinMoin import search, version
from MoinMoin import wikiutil
import calendar
import codecs
import datetime
import time
import re
import bisect
import operator
import urllib

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

try:
    set
except NameError:
    from sets import Set as set

try:
    import pytz
except ImportError:
    pytz = None

try:
    import vCalendar
except ImportError:
    vCalendar = None

escape = wikiutil.escape

__version__ = "0.8"

# Date labels.

month_labels = ["January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"]
weekday_labels = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

# Regular expressions where MoinMoin does not provide the required support.

category_regexp = None

# Page parsing.

definition_list_regexp = re.compile(ur'(?P<wholeterm>^(?P<optcomment>#*)\s+(?P<term>.*?):: )(?P<desc>.*?)$', re.UNICODE | re.MULTILINE)
category_membership_regexp = re.compile(ur"^\s*(?:(Category\S+)(?:\s+(Category\S+))*)\s*$", re.MULTILINE | re.UNICODE)

# Value parsing.

country_code_regexp = re.compile(ur'(?:^|\W)(?P<code>[A-Z]{2})(?:$|\W+$)', re.UNICODE)
location_normalised_regexp = re.compile(
    ur"(?:\d+\w*\s+)?"          # preceding postcode (optional)
    ur"(?P<location>"           # start of group of interest
    ur"\w[\w\s-]+?"             # area or town
    ur"(?:,(?:\s*[\w-]+)+)?"    # country (optional)
    ur")$", re.UNICODE)

# Month, date, time and datetime parsing.

month_regexp_str = ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})'
date_regexp_str = ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
time_regexp_str = ur'(?P<hour>[0-2][0-9]):(?P<minute>[0-5][0-9])(?::(?P<second>[0-6][0-9]))?'
timezone_offset_str = ur'(?P<offset>(UTC)?(?:(?P<sign>[-+])(?P<hours>[0-9]{2})(?::?(?P<minutes>[0-9]{2}))?))'
timezone_olson_str = ur'(?P<olson>[a-zA-Z]+(?:/[-_a-zA-Z]+){1,2})'
timezone_utc_str = ur'UTC'
timezone_regexp_str = ur'(?P<zone>' + timezone_offset_str + '|' + timezone_olson_str + '|' + timezone_utc_str + ')'
datetime_regexp_str = date_regexp_str + ur'(?:\s+' + time_regexp_str + ur'(?:\s+' + timezone_regexp_str + ur')?)?'

month_regexp = re.compile(month_regexp_str, re.UNICODE)
date_regexp = re.compile(date_regexp_str, re.UNICODE)
time_regexp = re.compile(time_regexp_str, re.UNICODE)
timezone_olson_regexp = re.compile(timezone_olson_str, re.UNICODE)
timezone_offset_regexp = re.compile(timezone_offset_str, re.UNICODE)
datetime_regexp = re.compile(datetime_regexp_str, re.UNICODE)

# iCalendar date and datetime parsing.

date_icalendar_regexp_str = ur'(?P<year>[0-9]{4})(?P<month>[0-9]{2})(?P<day>[0-9]{2})'
datetime_icalendar_regexp_str = date_icalendar_regexp_str + \
    ur'(?:' \
    ur'T(?P<hour>[0-2][0-9])(?P<minute>[0-5][0-9])(?P<second>[0-6][0-9])' \
    ur'(?P<utc>Z)?' \
    ur')?'

date_icalendar_regexp = re.compile(date_icalendar_regexp_str, re.UNICODE)
datetime_icalendar_regexp = re.compile(datetime_icalendar_regexp_str, re.UNICODE)

# Content type parsing.

encoding_regexp_str = ur'charset=(?P<encoding>[-A-Za-z0-9]+)'
encoding_regexp = re.compile(encoding_regexp_str)

# Simple content parsing.

verbatim_regexp = re.compile(ur'(?:'
    ur'<<Verbatim\((?P<verbatim>.*?)\)>>'
    ur'|'
    ur'\[\[Verbatim\((?P<verbatim2>.*?)\)\]\]'
    ur'|'
    ur'!(?P<verbatim3>.*?)(\s|$)?'
    ur'|'
    ur'`(?P<monospace>.*?)`'
    ur'|'
    ur'{{{(?P<preformatted>.*?)}}}'
    ur')', re.UNICODE)

# Utility functions.

def getCategoryPattern(request):
    global category_regexp

    try:
        return request.cfg.cache.page_category_regexact
    except AttributeError:

        # Use regular expression from MoinMoin 1.7.1 otherwise.

        if category_regexp is None:
            category_regexp = re.compile(u'^%s$' % ur'(?P<all>Category(?P<key>(?!Template)\S+))', re.UNICODE)
        return category_regexp

def getWikiDict(pagename, request):
    if Page(request, pagename).exists() and request.user.may.read(pagename):
        if hasattr(request.dicts, "dict"):
            return request.dicts.dict(pagename)
        else:
            return request.dicts[pagename]
    else:
        return None

def getContentEncoding(content_type):
    m = encoding_regexp.search(content_type)
    if m:
        return m.group("encoding")
    else:
        return None

def int_or_none(x):
    if x is None:
        return x
    else:
        return int(x)

def to_list(s, sep):
    return [x.strip() for x in s.split(sep) if x.strip()]

def sort_none_first(x, y):
    if x is None:
        return -1
    elif y is None:
        return 1
    else:
        return cmp(x, y)

def sort_start_first(x, y):
    x_ts = x.as_limits()
    if x_ts is not None:
        x_start, x_end = x_ts
        y_ts = y.as_limits()
        if y_ts is not None:
            y_start, y_end = y_ts
            start_order = cmp(x_start, y_start)
            if start_order == 0:
                return cmp(x_end, y_end)
            else:
                return start_order
    return 0

def sign(x):
    if x < 0:
        return -1
    else:
        return 1

# Utility classes and associated functions.

class Form:

    """
    A wrapper preserving MoinMoin 1.8.x (and earlier) behaviour in a 1.9.x
    environment.
    """

    def __init__(self, form):
        self.form = form

    def has_key(self, name):
        return not not self.form.getlist(name)

    def get(self, name, default=None):
        values = self.form.getlist(name)
        if not values:
            return default
        else:
            return values

    def __getitem__(self, name):
        return self.form.getlist(name)

class ActionSupport:

    """
    Work around disruptive MoinMoin changes in 1.9, and also provide useful
    convenience methods.
    """

    def get_form(self):
        return get_form(self.request)

    def _get_selected(self, value, input_value):

        """
        Return the HTML attribute text indicating selection of an option (or
        otherwise) if 'value' matches 'input_value'.
        """

        return input_value is not None and value == input_value and 'selected="selected"' or ''

    def _get_selected_for_list(self, value, input_values):

        """
        Return the HTML attribute text indicating selection of an option (or
        otherwise) if 'value' matches one of the 'input_values'.
        """

        return value in input_values and 'selected="selected"' or ''

    def _get_input(self, form, name, default=None):

        """
        Return the input from 'form' having the given 'name', returning either
        the input converted to an integer or the given 'default' (optional, None
        if not specified).
        """

        value = form.get(name, [None])[0]
        if not value: # true if 0 obtained
            return default
        else:
            return int(value)

    def get_month_lists(self, default_as_current=0):

        """
        Return two lists of HTML element definitions corresponding to the start
        and end month selection controls, with months selected according to any
        values that have been specified via request parameters.
        """

        _ = self._
        form = self.get_form()

        # Initialise month lists.

        start_month_list = []
        end_month_list = []

        start_month = self._get_input(form, "start-month", default_as_current and getCurrentMonth().month() or None)
        end_month = self._get_input(form, "end-month", start_month)

        # Prepare month lists, selecting specified months.

        if not default_as_current:
            start_month_list.append('<option value=""></option>')
            end_month_list.append('<option value=""></option>')

        for month in range(1, 13):
            month_label = escape(_(getMonthLabel(month)))
            selected = self._get_selected(month, start_month)
            start_month_list.append('<option value="%02d" %s>%s</option>' % (month, selected, month_label))
            selected = self._get_selected(month, end_month)
            end_month_list.append('<option value="%02d" %s>%s</option>' % (month, selected, month_label))

        return start_month_list, end_month_list

    def get_year_defaults(self, default_as_current=0):

        "Return defaults for the start and end years."

        form = self.get_form()

        start_year_default = form.get("start-year", [default_as_current and getCurrentYear() or ""])[0]
        end_year_default = form.get("end-year", [default_as_current and start_year_default or ""])[0]

        return start_year_default, end_year_default

    def get_day_defaults(self, default_as_current=0):

        "Return defaults for the start and end days."

        form = self.get_form()

        start_day_default = form.get("start-day", [default_as_current and getCurrentDate().day() or ""])[0]
        end_day_default = form.get("end-day", [default_as_current and start_day_default or ""])[0]

        return start_day_default, end_day_default

def get_form(request):

    "Work around disruptive MoinMoin changes in 1.9."

    if hasattr(request, "values"):
        return Form(request.values)
    else:
        return request.form

class send_headers_cls:

    """
    A wrapper to preserve MoinMoin 1.8.x (and earlier) request behaviour in a
    1.9.x environment.
    """

    def __init__(self, request):
        self.request = request

    def __call__(self, headers):
        for header in headers:
            parts = header.split(":")
            self.request.headers.add(parts[0], ":".join(parts[1:]))

def get_send_headers(request):

    "Return a function that can send response headers."

    if hasattr(request, "http_headers"):
        return request.http_headers
    elif hasattr(request, "emit_http_headers"):
        return request.emit_http_headers
    else:
        return send_headers_cls(request)

def escattr(s):
    return escape(s, 1)

# Textual representations.

def getSimpleWikiText(text):

    """
    Return the plain text representation of the given 'text' which may employ
    certain Wiki syntax features, such as those providing verbatim or monospaced
    text.
    """

    # NOTE: Re-implementing support for verbatim text and linking avoidance.

    return "".join([s for s in verbatim_regexp.split(text) if s is not None])

def getEncodedWikiText(text):

    "Encode the given 'text' in a verbatim representation."

    return "<<Verbatim(%s)>>" % text

def getPrettyTitle(title):

    "Return a nicely formatted version of the given 'title'."

    return title.replace("_", " ").replace("/", u"  ")

def getMonthLabel(month):

    "Return an unlocalised label for the given 'month'."

    return month_labels[month - 1] # zero-based labels

def getDayLabel(weekday):

    "Return an unlocalised label for the given 'weekday'."

    return weekday_labels[weekday]

def getNormalisedLocation(location):

    """
    Attempt to return a normalised 'location' of the form "<town>, <country>" or
    "<town>".
    """

    match = location_normalised_regexp.search(location)
    if match:
        return match.group("location")
    else:
        return None

def getLocationPosition(location, locations):

    """
    Attempt to return the position of the given 'location' using the 'locations'
    dictionary provided. If no position can be found, return a latitude of None
    and a longitude of None.
    """

    latitude, longitude = None, None

    if location is not None:
        try:
            latitude, longitude = map(getMapReference, locations[location].split())
        except (KeyError, ValueError):
            pass

    return latitude, longitude

# Action support functions.

def getPageRevision(page):

    "Return the revision details dictionary for the given 'page'."

    # From Page.edit_info...

    if hasattr(page, "editlog_entry"):
        line = page.editlog_entry()
    else:
        line = page._last_edited(page.request) # MoinMoin 1.5.x and 1.6.x

    # Similar to Page.mtime_usecs behaviour...

    if line:
        timestamp = line.ed_time_usecs
        mtime = wikiutil.version2timestamp(long(timestamp)) # must be long for py 2.2.x
        comment = line.comment
    else:
        mtime = 0
        comment = ""

    # Leave the time zone empty.

    return {"timestamp" : DateTime(time.gmtime(mtime)[:6] + (None,)), "comment" : comment}

# Category discovery and searching.

def getCategories(request):

    """
    From the AdvancedSearch macro, return a list of category page names using
    the given 'request'.
    """

    # This will return all pages with "Category" in the title.

    cat_filter = getCategoryPattern(request).search
    return request.rootpage.getPageList(filter=cat_filter)

def getCategoryMapping(category_pagenames, request):

    """
    For the given 'category_pagenames' return a list of tuples of the form
    (category name, category page name) using the given 'request'.
    """

    cat_pattern = getCategoryPattern(request)
    mapping = []
    for pagename in category_pagenames:
        name = cat_pattern.match(pagename).group("key")
        if name != "Category":
            mapping.append((name, pagename))
    mapping.sort()
    return mapping

def getCategoryPages(pagename, request):

    """
    Return the pages associated with the given category 'pagename' using the
    'request'.
    """

    query = search.QueryParser().parse_query('category:%s' % pagename)
    results = search.searchPages(request, query, "page_name")

    cat_pattern = getCategoryPattern(request)
    pages = []
    for page in results.hits:
        if not cat_pattern.match(page.page_name):
            pages.append(page)
    return pages

def getAllCategoryPages(category_names, request):

    """
    Return all pages belonging to the categories having the given
    'category_names', using the given 'request'.
    """

    pages = []
    pagenames = set()

    for category_name in category_names:

        # Get the pages and page names in the category.

        pages_in_category = getCategoryPages(category_name, request)

        # Visit each page in the category.

        for page_in_category in pages_in_category:
            pagename = page_in_category.page_name

            # Only process each page once.

            if pagename in pagenames:
                continue
            else:
                pagenames.add(pagename)

            pages.append(page_in_category)

    return pages

def getPagesFromResults(result_pages, request):

    "Return genuine pages for the given 'result_pages' using the 'request'."

    return [Page(request, page.page_name) for page in result_pages]

# Interfaces.

class ActsAsTimespan:
    pass

# Event resources providing collections of events.

class EventResource:

    "A resource providing event information."

    def __init__(self, url):
        self.url = url

    def getPageURL(self):

        "Return the URL of this page."

        return self.url

    def getFormat(self):

        "Get the format used by this resource."

        return "plain"

    def getMetadata(self):

        """
        Return a dictionary containing items describing the page's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        return {}

    def getEvents(self):

        "Return a list of events from this resource."

        return []

    def linkToPage(self, request, text, query_string=None):

        """
        Using 'request', return a link to this page with the given link 'text'
        and optional 'query_string'.
        """

        return linkToResource(self.url, request, text, query_string)

    # Formatting-related functions.

    def formatText(self, text, request, fmt):

        """
        Format the given 'text' using the specified 'request' and formatter
        'fmt'.
        """

        # Assume plain text which is then formatted appropriately.

        return fmt.text(text)

class EventCalendar(EventResource):

    "An iCalendar resource."

    def __init__(self, url, calendar):
        EventResource.__init__(self, url)
        self.calendar = calendar
        self.events = None

    def getEvents(self):

        "Return a list of events from this resource."

        if self.events is None:
            self.events = []

            _calendar, _empty, calendar = self.calendar

            for objtype, attrs, obj in calendar:

                # Read events.

                if objtype == "VEVENT":
                    details = {}

                    for property, attrs, value in obj:

                        # Convert dates.

                        if property in ("DTSTART", "DTEND", "CREATED", "DTSTAMP", "LAST-MODIFIED"):
                            if property in ("DTSTART", "DTEND"):
                                property = property[2:]
                            if attrs.get("VALUE") == "DATE":
                                value = getDateFromCalendar(value)
                            else:
                                value = getDateTimeFromCalendar(value)

                        # Convert numeric data.

                        elif property == "SEQUENCE":
                            value = int(value)

                        # Convert lists.

                        elif property == "CATEGORIES":
                            value = to_list(value, ",")

                        # Convert positions (using decimal values).

                        elif property == "GEO":
                            try:
                                value = map(getMapReferenceFromDecimal, to_list(value, ";"))
                                if len(value) != 2:
                                    continue
                            except (KeyError, ValueError):
                                continue

                        # Accept other textual data as it is.

                        elif property in ("LOCATION", "SUMMARY", "URL"):
                            pass

                        # Ignore other properties.

                        else:
                            continue

                        property = property.lower()
                        details[property] = value

                    self.events.append(CalendarEvent(self, details))

        return self.events

class EventPage:

    "An event page acting as an event resource."

    def __init__(self, page):
        self.page = page
        self.events = None
        self.body = None
        self.categories = None
        self.metadata = None

    def copyPage(self, page):

        "Copy the body of the given 'page'."

        self.body = page.getBody()

    def getPageURL(self):

        "Return the URL of this page."

        request = self.page.request
        return request.getQualifiedURL(self.page.url(request, relative=0))

    def getFormat(self):

        "Get the format used on this page."

        return self.page.pi["format"]

    def getMetadata(self):

        """
        Return a dictionary containing items describing the page's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        request = self.page.request

        # Get the initial revision of the page.

        revisions = self.getRevisions()
        event_page_initial = Page(request, self.getPageName(), rev=revisions[-1])

        # Get the created and last modified times.

        initial_revision = getPageRevision(event_page_initial)

        if self.metadata is None:
            self.metadata = {}
            self.metadata["created"] = initial_revision["timestamp"]
            latest_revision = self.getPageRevision()
            self.metadata["last-modified"] = latest_revision["timestamp"]
            self.metadata["sequence"] = len(revisions) - 1
            self.metadata["last-comment"] = latest_revision["comment"]

        return self.metadata

    def getRevisions(self):

        "Return a list of page revisions."

        return self.page.getRevList()

    def getPageRevision(self):

        "Return the revision details dictionary for this page."

        return getPageRevision(self.page)

    def getPageName(self):

        "Return the page name."

        return self.page.page_name

    def getPrettyPageName(self):

        "Return a nicely formatted title/name for this page."

        return getPrettyPageName(self.page)

    def getBody(self):

        "Get the current page body."

        if self.body is None:
            self.body = self.page.get_raw_body()
        return self.body

    def getEvents(self):

        "Return a list of events from this page."

        if self.events is None:
            details = {}
            self.events = [Event(self, details)]

            if self.getFormat() == "wiki":
                for match in definition_list_regexp.finditer(self.getBody()):

                    # Skip commented-out items.

                    if match.group("optcomment"):
                        continue

                    # Permit case-insensitive list terms.

                    term = match.group("term").lower()
                    desc = match.group("desc")

                    # Special value type handling.

                    # Dates.

                    if term in ("start", "end"):
                        desc = getDateTime(desc)

                    # Lists (whose elements may be quoted).

                    elif term in ("topics", "categories"):
                        desc = map(getSimpleWikiText, to_list(desc, ","))

                    # Position details.

                    elif term == "geo":
                        try:
                            desc = map(getMapReference, to_list(desc, None))
                            if len(desc) != 2:
                                continue
                        except (KeyError, ValueError):
                            continue

                    # Labels which may well be quoted.

                    elif term in ("title", "summary", "description", "location"):
                        desc = getSimpleWikiText(desc.strip())

                    if desc is not None:

                        # Handle apparent duplicates by creating a new set of
                        # details.

                        if details.has_key(term):

                            # Make a new event.

                            details = {}
                            self.events.append(Event(self, details))

                        details[term] = desc

        return self.events

    def setEvents(self, events):

        "Set the given 'events' on this page."

        self.events = events

    def getCategoryMembership(self):

        "Get the category names from this page."

        if self.categories is None:
            body = self.getBody()
            match = category_membership_regexp.search(body)
            self.categories = match and [x for x in match.groups() if x] or []

        return self.categories

    def setCategoryMembership(self, category_names):

        """
        Set the category membership for the page using the specified
        'category_names'.
        """

        self.categories = category_names

    def flushEventDetails(self):

        "Flush the current event details to this page's body text."

        new_body_parts = []
        end_of_last_match = 0
        body = self.getBody()

        events = iter(self.getEvents())

        event = events.next()
        event_details = event.getDetails()
        replaced_terms = set()

        for match in definition_list_regexp.finditer(body):

            # Permit case-insensitive list terms.

            term = match.group("term").lower()
            desc = match.group("desc")

            # Check that the term has not already been substituted. If so,
            # get the next event.

            if term in replaced_terms:
                try:
                    event = events.next()

                # No more events.

                except StopIteration:
                    break

                event_details = event.getDetails()
                replaced_terms = set()

            # Add preceding text to the new body.

            new_body_parts.append(body[end_of_last_match:match.start()])

            # Get the matching regions, adding the term to the new body.

            new_body_parts.append(match.group("wholeterm"))

            # Special value type handling.

            if event_details.has_key(term):

                # Dates.

                if term in ("start", "end"):
                    desc = desc.replace("YYYY-MM-DD", str(event_details[term]))

                # Lists (whose elements may be quoted).

                elif term in ("topics", "categories"):
                    desc = ", ".join([getEncodedWikiText(item) for item in event_details[term]])

                # Labels which must be quoted.

                elif term in ("title", "summary"):
                    desc = getEncodedWikiText(event_details[term])

                # Position details.

                elif term == "geo":
                    desc = " ".join(map(str, event_details[term]))

                # Text which need not be quoted, but it will be Wiki text.

                elif term in ("description", "link", "location"):
                    desc = event_details[term]

                replaced_terms.add(term)

            # Add the replaced value.

            new_body_parts.append(desc)

            # Remember where in the page has been processed.

            end_of_last_match = match.end()

        # Write the rest of the page.

        new_body_parts.append(body[end_of_last_match:])

        self.body = "".join(new_body_parts)

    def flushCategoryMembership(self):

        "Flush the category membership to the page body."

        body = self.getBody()
        category_names = self.getCategoryMembership()
        match = category_membership_regexp.search(body)

        if match:
            self.body = "".join([body[:match.start()], " ".join(category_names), body[match.end():]])

    def saveChanges(self):

        "Save changes to the event."

        self.flushEventDetails()
        self.flushCategoryMembership()
        self.page.saveText(self.getBody(), 0)

    def linkToPage(self, request, text, query_string=None):

        """
        Using 'request', return a link to this page with the given link 'text'
        and optional 'query_string'.
        """

        return linkToPage(request, self.page, text, query_string)

    # Formatting-related functions.

    def getParserClass(self, request, format):

        """
        Return a parser class using the 'request' for the given 'format', returning
        a plain text parser if no parser can be found for the specified 'format'.
        """

        try:
            return wikiutil.searchAndImportPlugin(request.cfg, "parser", format or "plain")
        except wikiutil.PluginMissingError:
            return wikiutil.searchAndImportPlugin(request.cfg, "parser", "plain")

    def formatText(self, text, request, fmt):

        """
        Format the given 'text' using the specified 'request' and formatter
        'fmt'.
        """

        fmt.page = self.page

        # Suppress line anchors.

        parser_cls = self.getParserClass(request, self.getFormat())
        parser = parser_cls(text, request, line_anchors=False)

        # Fix lists by indicating that a paragraph is already started.

        return request.redirectedOutput(parser.format, fmt, inhibit_p=True)

# Event details.

class Event(ActsAsTimespan):

    "A description of an event."

    def __init__(self, page, details):
        self.page = page
        self.details = details

        # Permit omission of the end of the event by duplicating the start.

        if self.details.has_key("start") and not self.details.get("end"):
            end = self.details["start"]

            # Make any end time refer to the day instead.

            if isinstance(end, DateTime):
                end = end.as_date()

            self.details["end"] = end

    def __repr__(self):
        return "<Event %r %r>" % (self.getSummary(), self.as_limits())

    def __hash__(self):

        """
        Return a dictionary hash, avoiding mistaken equality of events in some
        situations (notably membership tests) by including the URL as well as
        the summary.
        """

        return hash(self.getSummary() + self.getEventURL())

    def getPage(self):

        "Return the page describing this event."

        return self.page

    def setPage(self, page):

        "Set the 'page' describing this event."

        self.page = page

    def getEventURL(self):

        "Return the URL of this event."

        return self.page.getPageURL()

    def linkToEvent(self, request, text, query_string=None):

        """
        Using 'request', return a link to this event with the given link 'text'
        and optional 'query_string'.
        """

        return self.page.linkToPage(request, text, query_string)

    def getMetadata(self):

        """
        Return a dictionary containing items describing the event's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        # Delegate this to the page.

        return self.page.getMetadata()

    def getSummary(self, event_parent=None):

        """
        Return either the given title or summary of the event according to the
        event details, or a summary made from using the pretty version of the
        page name.

        If the optional 'event_parent' is specified, any page beneath the given
        'event_parent' page in the page hierarchy will omit this parent information
        if its name is used as the summary.
        """

        event_details = self.details

        if event_details.has_key("title"):
            return event_details["title"]
        elif event_details.has_key("summary"):
            return event_details["summary"]
        else:
            # If appropriate, remove the parent details and "/" character.

            title = self.page.getPageName()

            if event_parent and title.startswith(event_parent):
                title = title[len(event_parent.rstrip("/")) + 1:]

            return getPrettyTitle(title)

    def getDetails(self):

        "Return the details for this event."

        return self.details

    def setDetails(self, event_details):

        "Set the 'event_details' for this event."

        self.details = event_details

    # Timespan-related methods.

    def __contains__(self, other):
        return self == other

    def __eq__(self, other):
        if isinstance(other, Event):
            return self.getSummary() == other.getSummary() and self.getEventURL() == other.getEventURL() and self._cmp(other)
        else:
            return self._cmp(other) == 0

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self._cmp(other) == -1

    def __le__(self, other):
        return self._cmp(other) in (-1, 0)

    def __gt__(self, other):
        return self._cmp(other) == 1

    def __ge__(self, other):
        return self._cmp(other) in (0, 1)

    def _cmp(self, other):

        "Compare this event to an 'other' event purely by their timespans."

        if isinstance(other, Event):
            return cmp(self.as_timespan(), other.as_timespan())
        else:
            return cmp(self.as_timespan(), other)

    def as_timespan(self):
        details = self.details
        if details.has_key("start") and details.has_key("end"):
            return Timespan(details["start"], details["end"])
        else:
            return None

    def as_limits(self):
        ts = self.as_timespan()
        return ts and ts.as_limits()

class CalendarEvent(Event):

    "An event from a remote calendar."

    def getEventURL(self):

        "Return the URL of this event."

        return self.details.get("url") or self.page.getPageURL()

    def linkToEvent(self, request, text, query_string=None):

        """
        Using 'request', return a link to this event with the given link 'text'
        and optional 'query_string'.
        """

        return linkToResource(self.getEventURL(), request, text, query_string)

    def getMetadata(self):

        """
        Return a dictionary containing items describing the event's "created"
        time, "last-modified" time, "sequence" (or revision number) and the
        "last-comment" made about the last edit.
        """

        return {
            "created" : self.details.get("created") or self.details["dtstamp"],
            "last-modified" : self.details.get("last-modified") or self.details["dtstamp"],
            "sequence" : self.details.get("sequence") or 0,
            "last-comment" : ""
            }

# Obtaining event containers and events from such containers.

def getEventPages(pages):

    "Return a list of events found on the given 'pages'."

    # Get real pages instead of result pages.

    return map(EventPage, pages)

def getAllEventSources(request):

    "Return all event sources defined in the Wiki using the 'request'."

    sources_page = getattr(request.cfg, "event_aggregator_sources_page", "EventSourcesDict")

    # Remote sources are accessed via dictionary page definitions.

    return getWikiDict(sources_page, request)

def getEventResources(sources, calendar_start, calendar_end, request):

    """
    Return resource objects for the given 'sources' using the given
    'calendar_start' and 'calendar_end' to parameterise requests to the sources,
    and the 'request' to access configuration settings in the Wiki.
    """

    sources_dict = getAllEventSources(request)
    if not sources_dict:
        return []

    # Use dates for the calendar limits.

    if isinstance(calendar_start, Date):
        pass
    elif isinstance(calendar_start, Month):
        calendar_start = calendar_start.as_date(1)

    if isinstance(calendar_end, Date):
        pass
    elif isinstance(calendar_end, Month):
        calendar_end = calendar_end.as_date(-1)

    resources = []

    for source in sources:
        try:
            url, format = sources_dict[source].split()
        except (KeyError, ValueError):
            pass
        else:
            # Prevent local file access.

            if url.startswith("file:"):
                continue

            # Parameterise the URL.
            # Where other parameters are used, care must be taken to encode them
            # properly.

            url = url.replace("{start}", urllib.quote_plus(calendar_start and str(calendar_start) or ""))
            url = url.replace("{end}", urllib.quote_plus(calendar_end and str(calendar_end) or ""))

            # Get a parser.

            if format == "ical" and vCalendar is not None:
                parser = vCalendar.parse
                resource_cls = EventCalendar
            else:
                continue

            # See if the URL is cached.

            cache_key = cache.key(request, content=url)
            cache_entry = caching.CacheEntry(request, "EventAggregator", cache_key, scope='wiki')

            # If no entry exists, or if the entry is older than a certain age
            # (5 minutes by default), create one with the response from the URL.

            now = time.time()
            mtime = cache_entry.mtime()
            max_cache_age = int(getattr(request.cfg, "event_aggregator_max_cache_age", "300"))

            # NOTE: The URL could be checked and the 'If-Modified-Since' header
            # NOTE: (see MoinMoin.action.pollsistersites) could be checked.

            if not cache_entry.exists() or now - mtime >= max_cache_age:

                # Access the remote data source.

                cache_entry.open(mode="w")
                f = urllib.urlopen(url)
                try:
                    cache_entry.write(url + "\n")
                    cache_entry.write((f.headers.get("content-type") or "") + "\n")
                    cache_entry.write(f.read())
                finally:
                    cache_entry.close()
                    f.close()

            # Open the cache entry and read it.

            cache_entry.open()
            try:
                data = cache_entry.read()
            finally:
                cache_entry.close()

            # Process the entry, parsing the content.

            f = StringIO(data)
            try:
                url = f.readline()
                encoding = getContentEncoding(f.readline())
                uf = codecs.getreader(encoding or "utf-8")(f)
                try:
                    resources.append(resource_cls(url, parser(uf)))
                finally:
                    uf.close()
            finally:
                f.close()

    return resources

def getEventsFromResources(resources):

    "Return a list of events supplied by the given event 'resources'."

    events = []

    for resource in resources:

        # Get all events described by the resource.

        for event in resource.getEvents():

            # Remember the event.

            events.append(event)

    return events

# Event filtering and limits.

def getEventsInPeriod(events, calendar_period):

    """
    Return a collection containing those of the given 'events' which occur
    within the given 'calendar_period'.
    """

    all_shown_events = []

    for event in events:

        # Test for the suitability of the event.

        if event.as_timespan() is not None:

            # Compare the dates to the requested calendar window, if any.

            if event in calendar_period:
                all_shown_events.append(event)

    return all_shown_events

def getEventLimits(events):

    "Return the earliest and latest of the given 'events'."

    earliest = None
    latest = None

    for event in events:

        # Test for the suitability of the event.

        if event.as_timespan() is not None:
            ts = event.as_timespan()
            if earliest is None or ts.start < earliest:
                earliest = ts.start
            if latest is None or ts.end > latest:
                latest = ts.end

    return earliest, latest

def setEventTimestamps(request, events):

    """
    Using 'request', set timestamp details in the details dictionary of each of
    the 'events'.

    Return the latest timestamp found.
    """

    latest = None

    for event in events:
        event_details = event.getDetails()

        # Populate the details with event metadata.

        event_details.update(event.getMetadata())

        if latest is None or latest < event_details["last-modified"]:
            latest = event_details["last-modified"]

    return latest

def getOrderedEvents(events):

    """
    Return a list with the given 'events' ordered according to their start and
    end dates.
    """

    ordered_events = events[:]
    ordered_events.sort()
    return ordered_events

def getCalendarPeriod(calendar_start, calendar_end):

    """
    Return a calendar period for the given 'calendar_start' and 'calendar_end'.
    These parameters can be given as None.
    """

    # Re-order the window, if appropriate.

    if calendar_start is not None and calendar_end is not None and calendar_start > calendar_end:
        calendar_start, calendar_end = calendar_end, calendar_start

    return Timespan(calendar_start, calendar_end)

def getConcretePeriod(calendar_start, calendar_end, earliest, latest, resolution):

    """
    From the requested 'calendar_start' and 'calendar_end', which may be None,
    indicating that no restriction is imposed on the period for each of the
    boundaries, use the 'earliest' and 'latest' event months to define a
    specific period of interest.
    """

    # Define the period as starting with any specified start month or the
    # earliest event known, ending with any specified end month or the latest
    # event known.

    first = calendar_start or earliest
    last = calendar_end or latest

    # If there is no range of months to show, perhaps because there are no
    # events in the requested period, and there was no start or end month
    # specified, show only the month indicated by the start or end of the
    # requested period. If all events were to be shown but none were found show
    # the current month.

    if resolution == "date":
        get_current = getCurrentDate
    else:
        get_current = getCurrentMonth

    if first is None:
        first = last or get_current()
    if last is None:
        last = first or get_current()

    if resolution == "month":
        first = first.as_month()
        last = last.as_month()

    # Permit "expiring" periods (where the start date approaches the end date).

    return min(first, last), last

def getCoverage(events, resolution="date"):

    """
    Determine the coverage of the given 'events', returning a collection of
    timespans, along with a dictionary mapping locations to collections of
    slots, where each slot contains a tuple of the form (timespans, events).
    """

    all_events = {}
    full_coverage = TimespanCollection(resolution)

    # Get event details.

    for event in events:
        event_details = event.getDetails()

        # Find the coverage of this period for the event.

        # For day views, each location has its own slot, but for month
        # views, all locations are pooled together since having separate
        # slots for each location can lead to poor usage of vertical space.

        if resolution == "datetime":
            event_location = event_details.get("location")
        else:
            event_location = None

        # Update the overall coverage.

        full_coverage.insert_in_order(event)

        # Add a new events list for a new location.
        # Locations can be unspecified, thus None refers to all unlocalised
        # events.

        if not all_events.has_key(event_location):
            all_events[event_location] = [TimespanCollection(resolution, [event])]

        # Try and fit the event into an events list.

        else:
            slot = all_events[event_location]

            for slot_events in slot:

                # Where the event does not overlap with the events in the
                # current collection, add it alongside these events.

                if not event in slot_events:
                    slot_events.insert_in_order(event)
                    break

            # Make a new element in the list if the event cannot be
            # marked alongside existing events.

            else:
                slot.append(TimespanCollection(resolution, [event]))

    return full_coverage, all_events

def getCoverageScale(coverage):

    """
    Return a scale for the given coverage so that the times involved are
    exposed. The scale consists of a list of non-overlapping timespans forming
    a contiguous period of time.
    """

    times = set()
    for timespan in coverage:
        start, end = timespan.as_limits()

        # Add either genuine times or dates converted to times.

        if isinstance(start, DateTime):
            times.add(start)
        else:
            times.add(start.as_start_of_day())

        if isinstance(end, DateTime):
            times.add(end)
        else:
            times.add(end.as_date().next_day())

    times = list(times)
    times.sort(cmp_dates_as_day_start)

    scale = []
    first = 1
    start = None
    for time in times:
        if not first:
            scale.append(Timespan(start, time))
        else:
            first = 0
        start = time

    return scale

# Date-related functions.

def cmp_dates_as_day_start(a, b):

    """
    Compare dates/datetimes 'a' and 'b' treating dates without time information
    as the earliest time in a particular day.
    """

    are_equal = a == b

    if are_equal:
        a2 = a.as_datetime_or_date()
        b2 = b.as_datetime_or_date()

        if isinstance(a2, Date) and isinstance(b2, DateTime):
            return -1
        elif isinstance(a2, DateTime) and isinstance(b2, Date):
            return 1

    return cmp(a, b)

class Convertible:

    "Support for converting temporal objects."

    def _get_converter(self, resolution):
        if resolution == "month":
            return lambda x: x and x.as_month()
        elif resolution == "date":
            return lambda x: x and x.as_date()
        elif resolution == "datetime":
            return lambda x: x and x.as_datetime_or_date()
        else:
            return lambda x: x

class Temporal(Convertible):

    "A simple temporal representation, common to dates and times."

    def __init__(self, data):
        self.data = list(data)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.data)

    def __hash__(self):
        return hash(self.as_tuple())

    def as_tuple(self):
        return tuple(self.data)

    def convert(self, resolution):
        return self._get_converter(resolution)(self)

    def __cmp__(self, other):

        """
        The result of comparing this instance with 'other' is derived from a
        comparison of the instances' date(time) data at the highest common
        resolution, meaning that if a date is compared to a datetime, the
        datetime will be considered as a date. Thus, a date and a datetime
        referring to the same date will be considered equal.
        """

        if not isinstance(other, Temporal):
            return NotImplemented
        else:
            data = self.as_tuple()
            other_data = other.as_tuple()
            length = min(len(data), len(other_data))
            return cmp(data[:length], other_data[:length])

    def __sub__(self, other):

        """
        Return the difference between this object and the 'other' object at the
        highest common accuracy of both objects.
        """

        if not isinstance(other, Temporal):
            return NotImplemented
        else:
            data = self.as_tuple()
            other_data = other.as_tuple()
            if len(data) < len(other_data):
                return len(self.until(other))
            else:
                return len(other.until(self))

    def _until(self, start, end, nextfn, prevfn):

        """
        Return a collection of units of time by starting from the given 'start'
        and stepping across intervening units until 'end' is reached, using the
        given 'nextfn' and 'prevfn' to step from one unit to the next.
        """

        current = start
        units = [current]
        if current < end:
            while current < end:
                current = nextfn(current)
                units.append(current)
        elif current > end:
            while current > end:
                current = prevfn(current)
                units.append(current)
        return units

    def ambiguous(self):

        "Only times can be ambiguous."

        return 0

class Month(Temporal):

    "A simple year-month representation."

    def __str__(self):
        return "%04d-%02d" % self.as_tuple()[:2]

    def as_datetime(self, day, hour, minute, second, zone):
        return DateTime(self.as_tuple() + (day, hour, minute, second, zone))

    def as_date(self, day):
        if day < 0:
            weekday, ndays = self.month_properties()
            day = ndays + 1 + day
        return Date(self.as_tuple() + (day,))

    def as_month(self):
        return self

    def year(self):
        return self.data[0]

    def month(self):
        return self.data[1]

    def month_properties(self):

        """
        Return the weekday of the 1st of the month, along with the number of
        days, as a tuple.
        """

        year, month = self.as_tuple()[:2]
        return calendar.monthrange(year, month)

    def month_update(self, n=1):

        "Return the month updated by 'n' months."

        year, month = self.as_tuple()[:2]
        return Month((year + (month - 1 + n) / 12, (month - 1 + n) % 12 + 1))

    update = month_update

    def next_month(self):

        "Return the month following this one."

        return self.month_update(1)

    next = next_month

    def previous_month(self):

        "Return the month preceding this one."

        return self.month_update(-1)

    previous = previous_month

    def months_until(self, end):

        "Return the collection of months from this month until 'end'."

        return self._until(self.as_month(), end.as_month(), Month.next_month, Month.previous_month)

    until = months_until

class Date(Month):

    "A simple year-month-day representation."

    def constrain(self):
        year, month, day = self.as_tuple()[:3]

        month = max(min(month, 12), 1)
        wd, last_day = calendar.monthrange(year, month)
        day = max(min(day, last_day), 1)

        self.data[1:3] = month, day

    def __str__(self):
        return "%04d-%02d-%02d" % self.as_tuple()[:3]

    def as_datetime(self, hour, minute, second, zone):
        return DateTime(self.as_tuple() + (hour, minute, second, zone))

    def as_start_of_day(self):
        return self.as_datetime(None, None, None, None)

    def as_date(self):
        return self

    def as_datetime_or_date(self):
        return self

    def as_month(self):
        return Month(self.data[:2])

    def day(self):
        return self.data[2]

    def day_update(self, n=1):

        "Return the month updated by 'n' days."

        delta = datetime.timedelta(n)
        dt = datetime.date(*self.as_tuple()[:3])
        dt_new = dt + delta
        return Date((dt_new.year, dt_new.month, dt_new.day))

    update = day_update

    def next_day(self):

        "Return the date following this one."

        year, month, day = self.as_tuple()[:3]
        _wd, end_day = calendar.monthrange(year, month)
        if day == end_day:
            if month == 12:
                return Date((year + 1, 1, 1))
            else:
                return Date((year, month + 1, 1))
        else:
            return Date((year, month, day + 1))

    next = next_day

    def previous_day(self):

        "Return the date preceding this one."

        year, month, day = self.as_tuple()[:3]
        if day == 1:
            if month == 1:
                return Date((year - 1, 12, 31))
            else:
                _wd, end_day = calendar.monthrange(year, month - 1)
                return Date((year, month - 1, end_day))
        else:
            return Date((year, month, day - 1))

    previous = previous_day

    def days_until(self, end):

        "Return the collection of days from this date until 'end'."

        return self._until(self.as_date(), end.as_date(), Date.next_day, Date.previous_day)

    until = days_until

class DateTime(Date):

    "A simple date plus time representation."

    def constrain(self):
        Date.constrain(self)

        hour, minute, second = self.as_tuple()[3:6]

        if self.has_time():
            hour = max(min(hour, 23), 0)
            minute = max(min(minute, 59), 0)

            if second is not None:
                second = max(min(second, 60), 0) # support leap seconds

            self.data[3:6] = hour, minute, second

    def __str__(self):
        return Date.__str__(self) + self.time_string()

    def time_string(self):
        if self.has_time():
            data = self.as_tuple()
            time_str = " %02d:%02d" % data[3:5]
            if data[5] is not None:
                time_str += ":%02d" % data[5]
            if data[6] is not None:
                time_str += " %s" % data[6]
            return time_str
        else:
            return ""

    def as_HTTP_datetime_string(self):
        weekday = calendar.weekday(*self.data[:3])
        return "%s, %02d %s %04d %02d:%02d:%02d GMT" % ((
            getDayLabel(weekday),
            self.data[2],
            getMonthLabel(self.data[1]),
            self.data[0]
            ) + tuple(self.data[3:6]))

    def as_datetime(self):
        return self

    def as_date(self):
        return Date(self.data[:3])

    def as_datetime_or_date(self):

        """
        Return a date for this datetime if fields are missing. Otherwise, return
        this datetime itself.
        """

        if not self.has_time():
            return self.as_date()
        else:
            return self

    def __cmp__(self, other):

        """
        The result of comparing this instance with 'other' is, if both instances
        are datetime instances, derived from a comparison of the datetimes
        converted to UTC. If one or both datetimes cannot be converted to UTC,
        the datetimes are compared using the basic temporal comparison which
        compares their raw time data.
        """

        this = self

        if this.has_time():
            if isinstance(other, DateTime):
                if other.has_time():
                    this_utc = this.to_utc()
                    other_utc = other.to_utc()
                    if this_utc is not None and other_utc is not None:
                        return cmp(this_utc.as_tuple(), other_utc.as_tuple())
                else:
                    other = other.padded()
        else:
            this = this.padded()

        return Date.__cmp__(this, other)

    def has_time(self):

        """
        Return whether this object has any time information. Objects without
        time information can refer to the very start of a day.
        """

        return self.data[3] is not None and self.data[4] is not None

    def time(self):
        return self.data[3:]

    def seconds(self):
        return self.data[5]

    def time_zone(self):
        return self.data[6]

    def set_time_zone(self, value):
        self.data[6] = value

    def padded(self, empty_value=0):

        """
        Return a datetime with missing fields defined as being the given
        'empty_value' or 0 if not specified.
        """

        data = []
        for x in self.data[:6]:
            if x is None:
                data.append(empty_value)
            else:
                data.append(x)

        data += self.data[6:]
        return DateTime(data)

    def to_utc(self):

        """
        Return this object converted to UTC, or None if such a conversion is not
        defined.
        """

        if not self.has_time():
            return None

        offset = self.utc_offset()
        if offset:
            hours, minutes = offset

            # Invert the offset to get the correction.

            hours, minutes = -hours, -minutes

            # Get the components.

            hour, minute, second, zone = self.time()
            date = self.as_date()

            # Add the minutes and hours.

            minute += minutes
            if minute < 0 or minute > 59:
                hour += minute / 60
                minute = minute % 60

            # NOTE: This makes various assumptions and probably would not work
            # NOTE: for general arithmetic.

            hour += hours
            if hour < 0:
                date = date.previous_day()
                hour += 24
            elif hour > 23:
                date = date.next_day()
                hour -= 24

            return date.as_datetime(hour, minute, second, "UTC")

        # Cannot convert.

        else:
            return None

    def utc_offset(self):

        "Return the UTC offset in hours and minutes."

        zone = self.time_zone()
        if not zone:
            return None

        # Support explicit UTC zones.

        if zone == "UTC":
            return 0, 0

        # Attempt to return a UTC offset where an explicit offset has been set.

        match = timezone_offset_regexp.match(zone)
        if match:
            if match.group("sign") == "-":
                sign = -1
            else:
                sign = 1

            hours = int(match.group("hours")) * sign
            minutes = int(match.group("minutes") or 0) * sign
            return hours, minutes

        # Attempt to handle Olson time zone identifiers.

        dt = self.as_olson_datetime()
        if dt:
            seconds = dt.utcoffset().seconds
            hours = seconds / 3600
            minutes = (seconds % 3600) / 60
            return hours, minutes

        # Otherwise return None.

        return None

    def olson_identifier(self):

        "Return the Olson identifier from any zone information."

        zone = self.time_zone()
        if not zone:
            return None

        # Attempt to match an identifier.

        match = timezone_olson_regexp.match(zone)
        if match:
            return match.group("olson")
        else:
            return None

    def _as_olson_datetime(self, hours=None):

        """
        Return a Python datetime object for this datetime interpreted using any
        Olson time zone identifier and the given 'hours' offset, raising one of
        the pytz exceptions in case of ambiguity.
        """

        olson = self.olson_identifier()
        if olson and pytz:
            tz = pytz.timezone(olson)
            data = self.padded().as_tuple()[:6]
            dt = datetime.datetime(*data)

            # With an hours offset, find a time probably in a previously
            # applicable time zone.

            if hours is not None:
                td = datetime.timedelta(0, hours * 3600)
                dt += td

            ldt = tz.localize(dt, None)

            # With an hours offset, adjust the time to define it within the
            # previously applicable time zone but at the presumably intended
            # position.

            if hours is not None:
                ldt -= td

            return ldt
        else:
            return None

    def as_olson_datetime(self):

        """
        Return a Python datetime object for this datetime interpreted using any
        Olson time zone identifier, choosing the time from the zone before the
        period of ambiguity.
        """

        try:
            return self._as_olson_datetime()
        except (pytz.UnknownTimeZoneError, pytz.AmbiguousTimeError):

            # Try again, using an earlier local time and then stepping forward
            # in the chosen zone.
            # NOTE: Four hours earlier seems reasonable.

            return self._as_olson_datetime(-4)

    def ambiguous(self):

        "Return whether the time is local and ambiguous."

        try:
            self._as_olson_datetime()
        except (pytz.UnknownTimeZoneError, pytz.AmbiguousTimeError):
            return 1

        return 0

class Timespan(ActsAsTimespan, Convertible):

    """
    A period of time which can be compared against others to check for overlaps.
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

        # NOTE: Should perhaps catch ambiguous time problems elsewhere.

        if self.ambiguous() and self.start is not None and self.end is not None and start > end:
            self.start, self.end = end, start

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.start, self.end)

    def __hash__(self):
        return hash((self.start, self.end))

    def as_timespan(self):
        return self

    def as_limits(self):
        return self.start, self.end

    def ambiguous(self):
        return self.start is not None and self.start.ambiguous() or self.end is not None and self.end.ambiguous()

    def convert(self, resolution):
        return Timespan(*map(self._get_converter(resolution), self.as_limits()))

    def is_before(self, a, b):

        """
        Return whether 'a' is before 'b'. Since the end datetime of one period
        may be the same as the start datetime of another period, and yet the
        first period is intended to be concluded by the end datetime and not
        overlap with the other period, a different test is employed for datetime
        comparisons.
        """

        # Datetimes without times can be equal to dates and be considered as
        # occurring before those dates. Generally, datetimes should not be
        # produced without time information as getDateTime converts such
        # datetimes to dates.

        if isinstance(a, DateTime) and (isinstance(b, DateTime) or not a.has_time()):
            return a <= b
        else:
            return a < b

    def __contains__(self, other):

        """
        This instance is considered to contain 'other' if one is not before or
        after the other. If this instance overlaps or coincides with 'other',
        then 'other' is regarded as belonging to this instance's time period.
        """

        return self == other

    def __cmp__(self, other):

        """
        Return whether this timespan occupies the same period of time as the
        'other'. Timespans are considered less than others if their end points
        precede the other's start point, and are considered greater than others
        if their start points follow the other's end point.
        """

        if isinstance(other, ActsAsTimespan):
            other = other.as_timespan()

            if self.end is not None and other.start is not None and self.is_before(self.end, other.start):
                return -1
            elif self.start is not None and other.end is not None and self.is_before(other.end, self.start):
                return 1
            else:
                return 0

        else:
            if self.end is not None and self.is_before(self.end, other):
                return -1
            elif self.start is not None and self.is_before(other, self.start):
                return 1
            else:
                return 0

class TimespanCollection:

    """
    A class providing a list-like interface supporting membership tests at a
    particular resolution in order to maintain a collection of non-overlapping
    timespans.
    """

    def __init__(self, resolution, values=None):
        self.resolution = resolution
        self.values = values or []

    def as_timespan(self):
        return Timespan(*self.as_limits())

    def as_limits(self):

        "Return the earliest and latest points in time for this collection."

        if not self.values:
            return None, None
        else:
            first, last = self.values[0], self.values[-1]
            if isinstance(first, ActsAsTimespan):
                first = first.as_timespan().start
            if isinstance(last, ActsAsTimespan):
                last = last.as_timespan().end
            return first, last

    def convert(self, value):
        if isinstance(value, ActsAsTimespan):
            ts = value.as_timespan()
            return ts and ts.convert(self.resolution)
        else:
            return value.convert(self.resolution)

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __getitem__(self, i):
        return self.values[i]

    def __setitem__(self, i, value):
        self.values[i] = value

    def __contains__(self, value):
        test_value = self.convert(value)
        return test_value in self.values

    def append(self, value):
        self.values.append(value)

    def insert(self, i, value):
        self.values.insert(i, value)

    def pop(self):
        return self.values.pop()

    def insert_in_order(self, value):
        bisect.insort_left(self, value)

def getCountry(s):

    "Find a country code in the given string 's'."

    match = country_code_regexp.search(s)

    if match:
        return match.group("code")
    else:
        return None

def getDate(s):

    "Parse the string 's', extracting and returning a date object."

    dt = getDateTime(s)
    if dt is not None:
        return dt.as_date()
    else:
        return None

def getDateTime(s):

    """
    Parse the string 's', extracting and returning a datetime object where time
    information has been given or a date object where time information is
    absent.
    """

    m = datetime_regexp.search(s)
    if m:
        groups = list(m.groups())

        # Convert date and time data to integer or None.

        return DateTime(map(int_or_none, groups[:6]) + [m.group("zone")]).as_datetime_or_date()
    else:
        return None

def getDateFromCalendar(s):

    """
    Parse the iCalendar format string 's', extracting and returning a date
    object.
    """

    dt = getDateTimeFromCalendar(s)
    if dt is not None:
        return dt.as_date()
    else:
        return None

def getDateTimeFromCalendar(s):

    """
    Parse the iCalendar format datetime string 's', extracting and returning a
    datetime object where time information has been given or a date object where
    time information is absent.
    """

    m = datetime_icalendar_regexp.search(s)
    if m:
        groups = list(m.groups())

        # Convert date and time data to integer or None.

        return DateTime(map(int_or_none, groups[:6]) + [m.group("utc") and "UTC" or None]).as_datetime_or_date()
    else:
        return None

def getDateStrings(s):

    "Parse the string 's', extracting and returning all date strings."

    start = 0
    m = date_regexp.search(s, start)
    l = []
    while m:
        l.append("-".join(m.groups()))
        m = date_regexp.search(s, m.end())
    return l

def getMonth(s):

    "Parse the string 's', extracting and returning a month object."

    m = month_regexp.search(s)
    if m:
        return Month(map(int, m.groups()))
    else:
        return None

def getCurrentDate():

    "Return the current date as a (year, month, day) tuple."

    today = datetime.date.today()
    return Date((today.year, today.month, today.day))

def getCurrentMonth():

    "Return the current month as a (year, month) tuple."

    today = datetime.date.today()
    return Month((today.year, today.month))

def getCurrentYear():

    "Return the current year."

    today = datetime.date.today()
    return today.year

# Location-related functions.

class Reference:

    "A map reference."

    def __init__(self, degrees, minutes=0, seconds=0):
        self.degrees = degrees
        self.minutes = minutes
        self.seconds = seconds

    def __repr__(self):
        return "Reference(%d, %d, %f)" % (self.degrees, self.minutes, self.seconds)

    def __str__(self):
        return "%d:%d:%f" % (self.degrees, self.minutes, self.seconds)

    def __add__(self, other):
        if not isinstance(other, Reference):
            return NotImplemented
        else:
            s = sign(self.degrees)
            o = sign(other.degrees)
            carry, seconds = adc(s * self.seconds, o * other.seconds)
            carry, minutes = adc(s * self.minutes, o * other.minutes + carry)
            return Reference(self.degrees + other.degrees + carry, minutes, seconds)

    def __sub__(self, other):
        if not isinstance(other, Reference):
            return NotImplemented
        else:
            return self.__add__(Reference(-other.degrees, other.minutes, other.seconds))

    def _compare(self, op, other):
        if not isinstance(other, Reference):
            return NotImplemented
        else:
            return op(self.to_degrees(), other.to_degrees())

    def __eq__(self, other):
        return self._compare(operator.eq, other)

    def __ne__(self, other):
        return self._compare(operator.ne, other)

    def __lt__(self, other):
        return self._compare(operator.lt, other)

    def __le__(self, other):
        return self._compare(operator.le, other)

    def __gt__(self, other):
        return self._compare(operator.gt, other)

    def __ge__(self, other):
        return self._compare(operator.ge, other)

    def to_degrees(self):
        return sign(self.degrees) * (abs(self.degrees) + self.minutes / 60.0 + self.seconds / 3600.0)

    def to_pixels(self, scale):
        return self.to_degrees() * scale

def adc(x, y):
    result = x + y
    return divmod(result, 60)

def getPositionForReference(latitude, longitude, map_y, map_x, map_x_scale, map_y_scale):
    return (longitude - map_x).to_pixels(map_x_scale), (latitude - map_y).to_pixels(map_y_scale)

def getPositionForCentrePoint(position, map_x_scale, map_y_scale):
    x, y = position
    return x - map_x_scale / 2.0, y - map_y_scale / 2.0

def getMapReference(value):

    "Return a map reference by parsing the given 'value'."

    if value.find(":") != -1:
        return getMapReferenceFromDMS(value)
    else:
        return getMapReferenceFromDecimal(value)

def getMapReferenceFromDMS(value):

    """
    Return a map reference by parsing the given 'value' expressed as degrees,
    minutes, seconds.
    """

    values = value.split(":")
    values = map(int, values[:2]) + map(float, values[2:3])
    return Reference(*values)

def getMapReferenceFromDecimal(value):

    "Return a map reference by parsing the given 'value' in decimal degrees."

    value = float(value)
    degrees, remainder = divmod(abs(value * 3600), 3600)
    minutes, seconds = divmod(remainder, 60) 
    return Reference(sign(value) * degrees, minutes, seconds)

# User interface functions.

def getParameter(request, name, default=None):

    """
    Using the given 'request', return the value of the parameter with the given
    'name', returning the optional 'default' (or None) if no value was supplied
    in the 'request'.
    """

    return get_form(request).get(name, [default])[0]

def getQualifiedParameter(request, calendar_name, argname, default=None):

    """
    Using the given 'request', 'calendar_name' and 'argname', retrieve the
    value of the qualified parameter, returning the optional 'default' (or None)
    if no value was supplied in the 'request'.
    """

    argname = getQualifiedParameterName(calendar_name, argname)
    return getParameter(request, argname, default)

def getQualifiedParameterName(calendar_name, argname):

    """
    Return the qualified parameter name using the given 'calendar_name' and
    'argname'.
    """

    if calendar_name is None:
        return argname
    else:
        return "%s-%s" % (calendar_name, argname)

def getParameterDate(arg):

    "Interpret 'arg', recognising keywords and simple arithmetic operations."

    n = None

    if arg is None:
        return None

    elif arg.startswith("current"):
        date = getCurrentDate()
        if len(arg) > 8:
            n = int(arg[7:])

    elif arg.startswith("yearstart"):
        date = Date((getCurrentYear(), 1, 1))
        if len(arg) > 10:
            n = int(arg[9:])

    elif arg.startswith("yearend"):
        date = Date((getCurrentYear(), 12, 31))
        if len(arg) > 8:
            n = int(arg[7:])

    else:
        date = getDate(arg)

    if n is not None:
        date = date.day_update(n)

    return date

def getParameterMonth(arg):

    "Interpret 'arg', recognising keywords and simple arithmetic operations."

    n = None

    if arg is None:
        return None

    elif arg.startswith("current"):
        date = getCurrentMonth()
        if len(arg) > 8:
            n = int(arg[7:])

    elif arg.startswith("yearstart"):
        date = Month((getCurrentYear(), 1))
        if len(arg) > 10:
            n = int(arg[9:])

    elif arg.startswith("yearend"):
        date = Month((getCurrentYear(), 12))
        if len(arg) > 8:
            n = int(arg[7:])

    else:
        date = getMonth(arg)

    if n is not None:
        date = date.month_update(n)

    return date

def getFormDate(request, calendar_name, argname):

    """
    Return the date from the 'request' for the calendar with the given
    'calendar_name' using the parameter having the given 'argname'.
    """

    arg = getQualifiedParameter(request, calendar_name, argname)
    return getParameterDate(arg)

def getFormMonth(request, calendar_name, argname):

    """
    Return the month from the 'request' for the calendar with the given
    'calendar_name' using the parameter having the given 'argname'.
    """

    arg = getQualifiedParameter(request, calendar_name, argname)
    return getParameterMonth(arg)

def getFormDateTriple(request, yeararg, montharg, dayarg):

    """
    Return the date from the 'request' for the calendar with the given
    'calendar_name' using the parameters having the given 'yeararg', 'montharg'
    and 'dayarg' names.
    """

    year = getParameter(request, yeararg)
    month = getParameter(request, montharg)
    day = getParameter(request, dayarg)
    if year and month and day:
        return Date((int(year), int(month), int(day)))
    else:
        return None

def getFormMonthPair(request, yeararg, montharg):

    """
    Return the month from the 'request' for the calendar with the given
    'calendar_name' using the parameters having the given 'yeararg' and
    'montharg' names.
    """

    year = getParameter(request, yeararg)
    month = getParameter(request, montharg)
    if year and month:
        return Month((int(year), int(month)))
    else:
        return None

def getFullDateLabel(request, date):

    """
    Return the full month plus year label using the given 'request' and
    'year_month'.
    """

    if not date:
        return ""

    _ = request.getText
    year, month, day = date.as_tuple()[:3]
    start_weekday, number_of_days = date.month_properties()
    weekday = (start_weekday + day - 1) % 7
    day_label = _(getDayLabel(weekday))
    month_label = _(getMonthLabel(month))
    return "%s %s %s %s" % (day_label, day, month_label, year)

def getFullMonthLabel(request, year_month):

    """
    Return the full month plus year label using the given 'request' and
    'year_month'.
    """

    if not year_month:
        return ""

    _ = request.getText
    year, month = year_month.as_tuple()[:2]
    month_label = _(getMonthLabel(month))
    return "%s %s" % (month_label, year)

# Page-related functions.

def getPrettyPageName(page):

    "Return a nicely formatted title/name for the given 'page'."

    title = page.split_title(force=1)
    return getPrettyTitle(title)

def linkToPage(request, page, text, query_string=None):

    """
    Using 'request', return a link to 'page' with the given link 'text' and
    optional 'query_string'.
    """

    text = wikiutil.escape(text)
    return page.link_to_raw(request, text, query_string)

def linkToResource(url, request, text, query_string=None):

    """
    Using 'request', return a link to 'url' with the given link 'text' and
    optional 'query_string'.
    """

    if query_string:
        query_string = wikiutil.makeQueryString(query_string)
        url = "%s?%s" % (url, query_string)

    formatter = request.page and getattr(request.page, "formatter", None) or request.html_formatter

    output = []
    output.append(formatter.url(1, url))
    output.append(formatter.text(text))
    output.append(formatter.url(0))
    return "".join(output)

def getFullPageName(parent, title):

    """
    Return a full page name from the given 'parent' page (can be empty or None)
    and 'title' (a simple page name).
    """

    if parent:
        return "%s/%s" % (parent.rstrip("/"), title)
    else:
        return title

def fillEventPageFromTemplate(template_page, new_page, event_details, category_pagenames):

    """
    Using the given 'template_page', complete the 'new_page' by copying the
    template and adding the given 'event_details' (a dictionary of event
    fields), setting also the 'category_pagenames' to define category
    membership.
    """

    event_page = EventPage(template_page)
    new_event_page = EventPage(new_page)
    new_event_page.copyPage(event_page)

    if new_event_page.getFormat() == "wiki":
        new_event = Event(new_event_page, event_details)
        new_event_page.setEvents([new_event])
        new_event_page.setCategoryMembership(category_pagenames)
        new_event_page.saveChanges()

# vim: tabstop=4 expandtab shiftwidth=4
